import aioserial
import asyncio
import serial

import logging

_LOGGER = logging.getLogger(__name__)

class MLDaliController:
    __instance__ = None

    def __init__(self, port, baudrate, parity, stopbits, bytesize, timeout):
        """ Constructor.
        """

        if MLDaliController.__instance__ is None:
            self._ser = aioserial.AioSerial(
                                        port = port,
                                        baudrate = baudrate,
                                        timeout = timeout,
                                        parity = parity,
                                        stopbits = stopbits,
                                        bytesize = bytesize
                                    )
            self._registry = {}
            MLDaliController.__instance__ = self
   
    @staticmethod
    def register(component, port = "COM4", 
                    baudrate = 9600, 
                    parity = serial.PARITY_NONE, 
                    stopbits = serial.STOPBITS_ONE, 
                    bytesize = serial.EIGHTBITS, 
                    timeout = None) -> 'MLDaliController':
        if not MLDaliController.__instance__:
            MLDaliController(port, baudrate, parity, stopbits, bytesize, timeout)
            asyncio.create_task(MLDaliController.__instance__.monitor())
        MLDaliController.__instance__._registry[(component.address*2)+1] = component
        return MLDaliController.__instance__


    def open(self):
        self._ser.open()
    
    def close(self):
        self._ser.close()
    
    async def monitor(self):
        _LOGGER.debug("Start Monitoring")
        cmd = bytes()
        while True:
            rx = await self._ser.read_async(1)
            _LOGGER.debug(f"Observed: {rx}")
            if rx == b'\x02' or rx == b'\x04':
                cmd = rx
            else:
                cmd += rx
            
            if len(cmd) == 3:
                address = int.from_bytes(cmd[1:2],'big')
                component = self._registry.get(address, None)
                if component:
                    component.status_update(cmd)

    async def read_byte(self):
        rx = await self._ser.read_async(3)
        return rx
    
    async def sendCmd(self, tx):
        await self._ser.write_async(tx)



from .controller import MLDaliController
from mldali import LIGHT_SWITCHED_ON, LIGHT_SWITCHED_OFF, UNKNOWN_EVENT
import time
import logging

_LOGGER = logging.getLogger(__name__)

class MLDaliLight():

    def __init__(self, address, port="COM4"):
        self.address = address
        self._controller = MLDaliController.register(self,port=port)
        self.is_on = False
        self._listeners = []
    
    async def turn_on(self):
        cmd = bytearray([0x01, (self.address*2)+1, 5, 0xC1])
        time.sleep(.1) #This delay ensures that the command doesn't collapse with other commands being send on the bus
        await self._controller.sendCmd(cmd)

    
    async def turn_off(self):
        cmd = bytearray([0x01, (self.address*2)+1, 0, 0xC1])
        time.sleep(.1)
        await self._controller.sendCmd(cmd)
    
    def status_update(self, rx):
        _LOGGER.debug(f"Component at address {self.address} received status_update: {rx}")
        event = UNKNOWN_EVENT
        if rx[2:3] == b'\x05':
            self.is_on = True
            event = LIGHT_SWITCHED_ON
        elif rx[2:3] == b'\x00':
            self.is_on = False
            event = LIGHT_SWITCHED_OFF
        
        for call in self._listeners:
            call(event)
    
    def registerEventListener(self, callable):
        self._listeners.append(callable)